# httpmod functions - extends net/http

module Net  #:nodoc:
  class HTTP    #:nodoc:

    # Sends an HTTPRequest object REQUEST to the HTTP proxy such as polipo, and attach all new streams to the circuit specified in torctrl_cir.
    # torctrl_cir = {:torctrl, :cirnum,:hop}
    # Tor config must already be set to allow leaky circuits and attaching streams to the nodes other than the 3rd node.
    # Similar to HTTP.request
    # When called with a block, yields an HTTPResponse object.
    # The body of this response will not have been read yet;
    # the caller can process it using HTTPResponse#read_body,
    # if desired.
    #
    # Returns a HTTPResponse object.
    #
    # This method never raises Net::* exceptions.
    # The method has been implemented, but a circuit have been built to successfullty attach a stream to it.
    #
    def request_tor(req,torctrl_cir, body = nil, &block) 
      unless started?
        start {
          req['connection'] ||= 'close'
          return request(req, body, &block)
        }
      end
      if proxy_user()
        req.proxy_basic_auth proxy_user(), proxy_pass() unless use_ssl?
      end
      req.set_body_internal body
      res = transport_request_tor(req, torctrl_cir, &block)
      if sspi_auth?(res)
        sspi_auth(req)
        res = transport_request_tor(req, torctrl_cir, &block)
      end
      res
    end

    # Sends an HTTPRequest object GET REQUEST to the HTTP proxy such as polipo, and attach all new streams to the circuit specified in torctrl_cir.
    # torctrl_cir = {:torctrl, :cirnum,:hop}
    # Tor config must already be set to allow leaky circuits and attaching streams to the nodes other than the 3rd node.
    # Similar to HTTP.request_get
    # The method has been implemented, but a circuit have been built to successfullty attach a stream to it.
    #
    def request_torget(path, torctrl_cir, initheader = nil, &block)
      request_tor(Get.new(path, initheader),torctrl_cir, &block)
    end

    private
    
    # This transports requests to the proxy specified in the HTTP request, and then attaches all new streams to a prespecified circuit.
    # It works behind the scene when making requests using HTTP.request_torget and HTTP.request_tor
    # Further development in progress.
    #    - Must create new stream if cirnum==0
    #    - Attempt to recreate the same circuit if the circuit is closed. support sending Circuit instance in the hash instead of just cirnum
    #    ...more work to be done probably be used to make a browser for Tor, with limited functionality. No flash/javascript/cookies to start with, so no tracking.
    #
    def transport_request_tor(req, torctrl_cir)  #:nodoc:
      # torctrl_cir = {:torctrl, :cirnum,:hop}
      tcontrl = torctrl_cir[:torctrl]
      cir_num = torctrl_cir[:cirnum]
      hop_count = torctrl_cir[:hop] ? torctrl_cir[:hop] : 0 

      begin_transport req
      req.exec @socket, @curr_http_version, edit_path(req.path)

      #Get all new streams since the request was sent
      b = tcontrl.newstreams
      #  attach all new streams to cir_num
      b.each{|k|
          tcontrl.attach_stream(k, cir_num,hop_count)
      }

      begin
        res = HTTPResponse.read_new(@socket)
      end while res.kind_of?(HTTPContinue)
        res.reading_body(@socket, req.response_body_permitted?) {
          yield res if block_given?
        }
        end_transport req, res
        res
      rescue => exception
        D "Conn close because of error #{exception}"
        @socket.close if @socket and not @socket.closed?
        raise exception
      end

   end
 end